/*
 * Subverse
 * Copyright (C) 2007 Saulo Popov Zambiasi
 *
 * Subverse is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Subverse is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: saulopz@gmail.com
 */

#include "Element.h"
#include "World.h"

Element::Element(World *w, string imageName, int t) {
	name = imageName;
	world = w;
	timer = 0;
	loadImages(imageName);
	walking = false;
	working = true;
	direction = DOWN;
	myImage = image[direction];
	posX = 0;
	posY = 0;
	type = t;
	visionRange = 3;
	vision = NULL;
	worldSecound = world->getTimer()->getSecound();
	maxTime = 10;
}

void Element::run() {
	int aux = world->getTimer()->getSecound();
	if (worldSecound != aux) {
		worldSecound = world->getTimer()->getSecound();
		incTimer();
	
		int temp = direction;
		if (isStep() && (walking || working)) temp += 1;
	
		BITMAP *bmp = image[temp];
		if (!bmp) bmp = world->images->get("void");
		myImage = bmp;
	}
}

void Element::loadImages(string fileName) {
	for (unsigned int i = 0; i < 8; i++) {
		string temp = fileName + intToString(i);
		image[i] = world->images->get(temp);
	}
}

bool Element::move(int d) {
	bool out = false;
	setDirection(d);
	if(world->move(this, d)) {
		walking = true;
		out = true;
	}
	resetTime();
	return out;
}

void Element::resetTime() {
	timer = 0;
}

void Element::incTimer() {
	timer++;
	if (timer > maxTime) {
		timer = 0;
		walking = false;
	}
}

bool Element::isStep() {
	if (timer < ((int) (maxTime / 2))) return true;
	else return false;
}

bool Element::isWalking() {
	return walking;
}

bool Element::isWorking() {
	return working;
}

void Element::setWalking(bool b) {
	walking = b;
}

void Element::setWorking(bool b) {
	working = b;
}

void Element::setMaxTime(int t) {
	maxTime = t;
}

int Element::getTime() {
	return timer;
}

int Element::getMaxTime() {
	return maxTime;
}

int Element::getDirection() {
	return direction;
}

int Element::getVisionRange() {
	return visionRange;
}

void Element::setVision(Vision* v) {
	vision = v;
}

Vision* Element::getVision() {
	return vision;
}

void Element::setDirection(int dir) {
	if ((dir >= 0) && (dir < 8)) {
		direction = dir;
	}
}

World* Element::getWorld() {
	return world;
}

void Element::setX(int x) {
	posX = x;
}

void Element::setY(int y) {
	posY = y;
}

int Element::getX() {
	return posX;
}

int Element::getY() {
	return posY;
}

string Element::getName() {
	return name;
}

int Element::getType() {
	return type;
}

BITMAP* Element::getImage() {
	return myImage;
}

Element::~Element() {
	for (unsigned int i = 0; i < 8; i++) {
		if (image[i] != NULL) image[i] = NULL;
	}
}
