/*
 * Subverse
 * Copyright (C) 2007 Saulo Popov Zambiasi
 *
 * Subverse is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Subverse is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: saulopz@gmail.com
 */

#include "World.h"

using namespace std;

World::World(std::string fileName) {
	background = create_bitmap(MAX_X + (CELL_SIZE * 2), MAX_Y + (CELL_SIZE * 2));
	clear_to_color(background, makecol(255, 0, 0));
	timer = new Timer();
	height = 0;
	width = 0;
	posX = 0;
	posY = 0;
	avatar = NULL;
	
	images = new Images();
	images->add("avatar");
	images->add("cyclope");
	images->add("water");
	images->add("adobe");
	images->add("spider");
	images->add("fire");
	images->add("lawn");
	images->add("littlelawn");
	images->add("rocks");
	images->add("tree");
	images->add("wall");
	
	loadWorld(fileName);
}

Element* World::get(Element* e, int type, int direction) {
	Element* aux = NULL;
	switch(direction) {
		case UP   :
			if (e->getY() > 0) {
				aux = get(type, e->getX(), e->getY()-1);
			}
			break;
		case DOWN :
			if (e->getY() < getHeight()-1) {
				aux = get(type, e->getX(), e->getY()+1);
			}
			break;
		case LEFT : 
			if (e->getX() > 0) {
				aux = get(type, e->getX()-1, e->getY());
			}
			break;
		case RIGHT:
			if (e->getX() < getWidth()-1) {
				aux = get(type, e->getX()+1, e->getY());
			}
			break;
	}
	return aux;
}

bool World::move(Element* e, int direction) {
	bool freeWay = false;
	Element* land = get(e, LAND, direction);
	if (land != NULL) {
		if (land->getType() < WALL) {
			freeWay = true;
		}
		Element *aux = get(e, ELEMENT, direction);
		if (aux != NULL) freeWay = false;
	}
	bool moved = false;
	if (freeWay) {
		int auxX = e->getX();
		int auxY = e->getY();
		switch(direction) {
			case UP    : auxY--; break;
			case DOWN  : auxY++; break;
			case RIGHT : auxX++; break;
			case LEFT  : auxX--; break;
		}
		set(ELEMENT, auxX, auxY, e);
		set(ELEMENT, e->getX(), e->getY(), NULL);
		e->setX(auxX);
		e->setY(auxY);
		moved = true;
	}
	return moved;
}

void World::draw() {
	show_mouse(NULL);
	//clear(background);
	// draw land
	for (int x = (-1); x < getScreenWidth()+1; x++) {
		for (int y = (-1); y < getScreenHeight()+1; y++) {
			Element* e = get(LAND, getX() + x, getY() + y);
			if (e != NULL) {
				draw_sprite(background, e->getImage(), (x+1) * CELL_SIZE, (y+1) * CELL_SIZE);
			}
		}
	}
	// draw elements
	for (int x = (-1); x < getScreenWidth()+1; x++) {
		for (int y = (-1); y < getScreenHeight()+1; y++) {
			Element* e = get(ELEMENT, getX() + x, getY() + y);
			if (e != NULL) {
				int auxX = (x+1) * CELL_SIZE;
				int auxY = (y+1) * CELL_SIZE;
				//int auxX = (e->getX() - getX()) * CELL_SIZE;
				//int auxY = (e->getY() - getY()) * CELL_SIZE;
				if (e->isWalking() && e->isStep()) {
					switch (e->getDirection()) {
						case DOWN  : auxY -= (int) (CELL_SIZE / 2); break;
						case RIGHT : auxX -= (int) (CELL_SIZE / 2); break;
						case UP    : auxY += (int) (CELL_SIZE / 2); break;
						case LEFT  : auxX += (int) (CELL_SIZE / 2); break;
					}
				}
				draw_sprite(background, e->getImage(), auxX, auxY);
			}
		}
	}
	show_mouse(background);
}

void World::run() {
	Element *e = NULL;
	while (!key[KEY_ESC]) {
		for (int x = 0; x < getWidth(); x++) {
			for (int y = 0; y < getHeight(); y++) {
				e = get(ELEMENT, x, y);
				if (e != NULL) e->run();
				e = get(LAND, x, y);
				if (e != NULL) e->run();
			}
		}
		setX(avatar->getX() - (int) (getScreenWidth() / 2));
		setY(avatar->getY() - (int) (getScreenHeight() / 2));
		incTimer();
		draw();
		int auxX = CELL_SIZE;
		int auxY = CELL_SIZE;
		if (avatar->isWalking() && avatar->isStep()) {
			switch (avatar->getDirection()) {
				case UP    :
					if ((getY() > 0) && (getY() < (getHeight() - getScreenHeight()))) { 
						auxY += (int) (CELL_SIZE / 2);
					}
					break;
				case DOWN  :
					if ((getY() > 0) && (getY() < (getHeight() - getScreenHeight()))) {
						auxY -= (int) (CELL_SIZE / 2);
					}
					break;
				case LEFT  :
					if ((getX() > 0) && (getX() < (getWidth() - getScreenWidth()))) {
						auxX += (int) (CELL_SIZE / 2);
					}
					break;
				case RIGHT :
					if ((getX() > 0) && (getX() < (getWidth() - getScreenWidth()))) {
						auxX -= (int) (CELL_SIZE / 2);
					}
					break;
			}
		}
		blit(getImage(), screen, auxX, auxY, 0, 0, MAX_X, MAX_Y);
		clear_keybuf();
		rest(50);
	}
}

void World::loadWorld(std::string fileName) {
	LoadWorld *loadWorld = new LoadWorld(fileName);
	height = loadWorld->getHeight();
	width = loadWorld->getWidth();
	
	Type* typeLand = new Type();
	Type* typeElement = new Type();
	
	for (int y = 0; y < getHeight(); y++) {
		Line* lineElement = new Line();
		Line* lineLand = new Line();
    	for (int x = 0; x < getWidth(); x++) {
    		Element *e;
    		int iten = loadWorld->getNext(ELEMENT);
    		switch (iten) {
    			case AVATAR :
    				if (avatar == NULL)	{
    					e = new Avatar(this);
    					avatar = e;
    					setX(x - (int) (getScreenWidth() / 2));
    					setY(y - (int) (getScreenHeight() / 2));
    				} else e = NULL;
    				break;
    			case CYCLOPE : e = new Cyclope(this); break;
    			case FIRE : e = new Element(this, "fire", FIRE); break;
    			case TREE : e = new Element(this, "tree", TREE); break;
    			case SPIDER : e = new Spider(this); break;
    			default : e = NULL; 
    		}
    		if (e != NULL) {
    			e->setX(x);
    			e->setY(y);
    		}
    		lineElement->x.push_back(e);
  			
    		iten = loadWorld->getNext(LAND);
    		switch (iten) {
    			case ADOBE : e = new Element(this, "adobe", ADOBE); break;
    			case WALL : e = new Element(this, "wall", WALL); break;
    			case GROUND : e = new Element(this, "ground", GROUND); break;
    			case ROCKS : e = new Element(this, "rocks", ROCKS); break;
    			case LAWN : e = new Element(this, "lawn", LAWN); break;
    			case LITTLELAWN : e = new Element(this, "littlelawn", LITTLELAWN); break;
    			case WATER : e = new Element(this, "water", WATER); break;
    			default : e = NULL;
    		}
    		if (e != NULL) {
    			e->setX(x);
    			e->setY(y);
    		}
    		lineLand->x.push_back(e);
    	}
    	typeElement->y.push_back(lineElement);
    	typeLand->y.push_back(lineLand);
    }
    element.push_back(typeLand);
    element.push_back(typeElement);
}

Vision* World::getVision(Element* e, int range) {
	Vision* vision = new Vision();
	
	if (e == NULL) return NULL;
	
	int elementX  = e->getX();
	int elementY  = e->getY();
	int direction = e->getDirection();

	for(int i = 0; i < range; i++) {
		for (int j = (-i); j <= i; j++) {
			int auxX = elementX;
			int auxY = elementY;
			switch(direction) {
				case UP :
					auxX += j;
					auxY -= i - 1;
					break;
				case DOWN :
					auxX += j;
					auxY += i + 1;
					break;
				case LEFT :
					auxX -= i - 1;
					auxY += j;
					break;
				case RIGHT :
					auxX += i + 1;
					auxY += j;
					break;
			}
			Element* element = get(ELEMENT, auxX, auxY);
			if (element != NULL) vision->element.push_back(element);
			element = get(LAND, auxX, auxY);
			if (element != NULL) vision->land.push_back(element);
		}
	}
	return vision;
}

Element* World::get(int type, int x, int y) {
	if ((x < 0) || (x >= getWidth()) ||
		(y < 0) || (y >= getHeight()) ||
		(type < 0) || (type > 1)) {
		return NULL;
	} else return element[type]->y[y]->x[x];
}

void World::set(int type, int x, int y, Element* e) {
	if ((x >= 0) && (x < getWidth()) &&
		(y >= 0) && (y < getHeight()) &&
		(type >= 0) && (type < 2)) {
		element[type]->y[y]->x[x] = e;
	}
}

int World::getWidth() {
	return width;
}

int World::getHeight() {
	return height;
}

int World::getX() {
	return posX;
}

int World::getY() {
	return posY;
}

int World::getScreenWidth() {
	return ((int) (MAX_X / CELL_SIZE));
}

int World::getScreenHeight() {
	return ((int) (MAX_Y / CELL_SIZE));
}

void World::setX(int in) {
	int temp = getScreenWidth();
	if (in > (width - temp)) {
		posX = width - temp;
	} else if (in < 0) {
		posX = 0;
	} else posX = in;
}

void World::setY(int in) {
	int temp = getScreenHeight();
	if (in > (height - temp)) {
		posY = height - temp;
	} else if (in < 0) {
		posY = 0;
	} else posY = in;
}

void World::setPosition(int x_in, int y_in) {
	setX(x_in);
	setY(y_in);
}

Timer *World::getTimer() {
	return timer;
}

BITMAP *World::getImage() {
	return background;
}

void World::incTimer() {
	timer->inc();
}

World::~World() {
	Element* e = NULL;
	for (int y = 0; y < getHeight(); y++) {
		for (int x = 0; x < getWidth(); x++) {
			e = get(ELEMENT, x, y);
			if (e != NULL) delete e;
			e = get(LAND, x, y);
			if (e != NULL) delete e;
		}
	}
	element.clear();
	delete images ;
}
